import json
from datetime import datetime
from typing import Union, Optional

import aiohttp

from aiocryptocurrency import TransactionSet, Transaction
from aiocryptocurrency.coins import Coin


class Firo(Coin):
    def __init__(self):
        super(Firo, self).__init__()
        self.host = '127.0.0.1'
        self.port = 8888
        self.basic_auth: Optional[tuple[str]] = None
        self.url = None

    async def send(self, address: str, amount: float) -> str:
        """returns txid"""
        if amount <= 0:
            raise Exception("amount cannot be zero or less")

        data = {
            "method": "sendtoaddress",
            "params": [address, amount]
        }

        blob = await self._make_request(data=data)
        return blob['result']

    async def create_address(self) -> dict:
        """returns an address"""
        data = {
            "method": "getnewaddress"
        }
        blob = await self._make_request(data=data)

        address = blob['result']
        if address is None or not isinstance(address, str):
            raise Exception("Invalid result")

        return {
            "address": address
        }

    async def tx_details(self, txid: str):
        if not isinstance(txid, str) or not txid:
            raise Exception("bad address")

        data = {
            "method": "gettransaction",
            "params": [txid]
        }

        blob = await self._make_request(data=data)
        return blob['result']

    async def list_txs(self, address: str, minimum_confirmations: int = 3) -> Optional[TransactionSet]:
        txset = TransactionSet()
        if not isinstance(address, str) or not address:
            raise Exception("bad address")

        results = await self._make_request(data={
            "method": "listreceivedbyaddress",
            "params": [minimum_confirmations]
        })

        if not isinstance(results.get('result'), list):
            return txset

        try:
            result = [r for r in results['result'] if r['address'] == address][0]
        except Exception as ex:
            return txset

        for txid in result.get('txids', []):
            # fetch tx details
            tx = await self.tx_details(txid)

            # fetch blockheight
            tx['blockheight'] = await self.blockheight(tx['blockhash'])
            date = datetime.fromtimestamp(tx['blocktime'])

            txset.add(Transaction(amount=tx['amount'],
                                  txid=tx['txid'],
                                  date=date,
                                  blockheight=tx['blockheight'],
                                  direction='in',
                                  confirmations=tx['confirmations']))

        return txset

    async def blockheight(self, blockhash: str) -> int:
        """blockhash -> blockheight"""
        if not isinstance(blockhash, str) or not blockhash:
            raise Exception("bad address")

        data = {
            "method": "getblock",
            "params": [blockhash]
        }
        blob = await self._make_request(data=data)

        height = blob['result'].get('height', 0)
        return height

    async def _generate_url(self) -> None:
        self.url = f'http://{self.host}:{self.port}/'

    async def _make_request(self, data: dict = None) -> dict:
        await self._generate_url()

        opts = {
            "headers": {
                "User-Agent": self.user_agent
            }
        }

        if self.basic_auth:
            opts['auth'] = await self._make_basic_auth()

        async with aiohttp.ClientSession(**opts) as session:
            async with session.post(self.url, json=data) as resp:
                if resp.status == 401:
                    raise Exception("Unauthorized")
                blob = await resp.json()
                if 'result' not in blob:
                    if blob:
                        blob = json.dumps(blob, indent=4, sort_keys=True)
                    raise Exception(f"Invalid response: {blob}")
                return blob



from setuptools import setup, find_packages

with open('README.md') as f:
    long_description = f.read()

INSTALL_REQUIRES = [
    "aiohttp",
    "twine"
]

setup(
    name='aiocryptocurrency',
    version='0.1.6',
    author='Sander',
    author_email='sander@sanderf.nl',
    python_requires='>=3.6',
    packages=['aiocryptocurrency'],
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=INSTALL_REQUIRES,
    url='https://github.com/sanderfoobar/aiocryptocurrency',
    description='Abstraction library for managing funds for various cryptocurrencies via their RPCs',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: BSD License",
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    keywords='cryptocurrency'
)



import json
from operator import attrgetter
from typing import List, Union, Optional
from datetime import datetime

from dateutil.parser import parse as dateutil_parse


class Transaction:
    def __init__(self, amount: float, txid: str, direction: str, date: Union[datetime, str], confirmations: int = 0, blockheight: int = None):
        self.amount: float = amount
        self.txid: str = txid
        if isinstance(date, str):
            self.date = dateutil_parse(date)
        else:
            self.date: datetime = date
        self.blockheight: int = blockheight
        self.direction: str = direction
        self.confirmations: int = confirmations

    @classmethod
    def from_json(cls, data: dict):
        return Transaction(
            data['amount'],
            data['txid'],
            datetime.strptime(data['d1ate'], '%Y-%m-%d %H:%M'),
            data['blockheight'],
            data['direction'],
            data['confirmations']
        )

    def json(self):
        return {
            'amount': self.amount,
            'txid': self.txid,
            'date': self.date.strftime('%Y-%m-%d %H:%M'),
            'blockheight': self.blockheight,
            'direction': self.direction,
            'confirmations': self.confirmations
        }


class TransactionSet:
    def __init__(self):
        self._transactions: list[Transaction] = []
        self._idx = 0

    def serialize(self) -> str:
        data = [tx.json() for tx in self._transactions]
        if not data:
            return json.dumps([])

        return json.dumps(data, sort_keys=True, indent=True)

    @classmethod
    def from_json(cls, val: dict):
        cls = TransactionSet()
        if not val or not isinstance(val, list):
            return cls

        for r in val:
            tx = Transaction(**r)
            cls.add(tx)
        return cls

    def __iter__(self):
        self._idx = 0
        return self

    def __next__(self):
        if self._idx < len(self._transactions):
            tx = self._transactions[self._idx]
            self._idx += 1
            return tx
        raise StopIteration

    def __len__(self):
        return len(self._transactions)

    def __getitem__(self, txid: str) -> Optional[Transaction]:
        for transaction in self._transactions:
            if transaction.txid == txid:
                return transaction

    def __add__(self, other: 'TransactionSet'):
        ts = TransactionSet()
        ts._transactions = self._transactions
        for tx in other:
            ts.add(tx)
        ts.sort()
        return ts

    def sort(self):
        self._transactions.sort(key=lambda tx: tx.date, reverse=True)

    def add(self, item: Transaction):
        self._transactions.append(item)
        self.sort()

    def filter(self, direction='in') -> 'TransactionSet':
        txset = TransactionSet()
        txset._transactions = [tx for tx in self._transactions if tx.direction == direction]
        return txset

    def sum(self, direction='in') -> float:
        try:
            return sum(tx.amount for tx in self._transactions if tx.direction == direction)
        except Exception as ex:
            return 0.0



import re
import os
from typing import Union, Optional

import aiohttp

from aiocryptocurrency import TransactionSet, Transaction

SUPPORTED_COINS = {
        "firo": "firo",
        "monero": "xmr",
        "wownero": "wow"
}


class Coin:
    def __init__(self):
        self.host = "127.0.0.1"
        self.port = 0
        self.basic_auth: Optional[tuple[str]] = None
        self.verify_tls = False
        self.user_agent = "aiocryptocurrency"

    async def send(self, address: str, amount: float):
        raise NotImplemented()

    async def create_address(self, *args, **kwargs):
        raise NotImplemented()

    async def list_txs(self, *args, **kwargs) -> TransactionSet:
        raise NotImplemented()

    async def _generate_url(self):
        raise NotImplemented()

    async def _make_basic_auth(self):
        return aiohttp.BasicAuth(
            login=self.basic_auth[0],
            password=self.basic_auth[1]
        )



from datetime import datetime

import aiohttp

from aiocryptocurrency import Transaction, TransactionSet
from aiocryptocurrency.coins import Coin


class Nero(Coin):
    # @TODO: support authentication
    def __init__(self):
        super(Nero, self).__init__()
        self.host = "127.0.0.1"
        self.url = None
        self._div = None

    async def send(self, address: str, amount: float) -> str:
        """returns txid"""
        await self._generate_url()
        if amount <= 0:
            raise Exception("amount cannot be zero or less")

        amount_atomic = int(amount * self._div)

        async with aiohttp.ClientSession() as session:
            data = {
                "jsonrpc": "2.0",
                "id": "0",
                "method": "transfer",
                "params": {
                    "destinations": [{
                        "amount": amount_atomic,
                        "address": address
                    }]
                }
            }

            async with session.post(self.url, json=data) as resp:
                blob = await resp.json()
                if 'result' not in blob:
                    raise Exception("Invalid response")
                return blob['result']['tx_hash']

    async def create_address(self) -> dict:
        """returns both payment_id and address"""
        await self._generate_url()

        async with aiohttp.ClientSession() as session:
            data = {
                "jsonrpc": "2.0",
                "id": "0",
                "method": "make_integrated_address"
            }

            async with session.get(self.url, json=data) as resp:
                blob = await resp.json()
                if 'result' not in blob:
                    raise Exception("Invalid response")

                res = blob['result']
                if "integrated_address" not in res or 'payment_id' not in res:
                    raise Exception("Invalid result response")

                return {
                    "address": res['integrated_address'],
                    "payment_id": res['payment_id']
                }

    async def list_txs(self, payment_id: str) -> TransactionSet:
        txset = TransactionSet()
        await self._generate_url()

        if not isinstance(payment_id, str) or not payment_id:
            raise Exception("bad payment_id")

        async with aiohttp.ClientSession() as session:
            data = {
                "jsonrpc": "2.0",
                "id": "0",
                "method": "get_payments",
                "params": {
                    "payment_id": payment_id
                }
            }

            async with session.get(self.url, json=data) as resp:
                blob = await resp.json()
                if 'result' not in blob:
                    raise Exception("Invalid response")

                res = blob['result']
                if not isinstance(res['payments'], list):
                    raise Exception("Invalid result response")

                transactions = res['payments']
                for transaction in transactions:
                    amount = float(transaction['amount']) / self._div

                    tx = Transaction(amount=amount,
                                     txid=transaction['tx_hash'],
                                     blockheight=transaction['block_height'],
                                     direction='in')
                    txset.add(tx)

        return txset

    async def _generate_url(self) -> None:
        self.url = f"http://{self.host}:{self.port}/json_rpc"


class Wownero(Nero):
    def __init__(self):
        # ./wownero-wallet-rpc --rpc-bind-port 45678 --disable-rpc-login --wallet-file /home/user/wallet_name.keys --password
        super(Wownero, self).__init__()
        self.port = 45678
        self._div = 1e11


class Monero(Nero):
    def __init__(self):
        super(Monero, self).__init__()
        self.port = 34567
        self._div = 1e12
